/**********
 * Copyright (c) 2003-2005 Greg Parker.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY GREG PARKER ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **********/

#include "ssh.h"
#include "rsrc.h"
#include "session.h"
#include "packetizer.h"
#include "transport.h"
#include "connection.h"
#include "formutils.h"
#include "forms/publickeychoiceform.h"
#include "forms/kbdintform.h"
#include "forms/kbdint0form.h"

#include "vt100/vt100.h"
#include "data/prefs.h"

// The "current session" being processed. 
// This is used for malloc(), printf(), etc when deep inside libraries.
// current_ss is set by the ssh_* functions below.
static ssh_session_t *current_ss = NULL;


static ssh_session_t *ssh_set_current(ssh_session_t *ss)
{
    ssh_session_t *old_ss = current_ss;
    current_ss = ss;
    return old_ss;
}


ssh_session_t *ssh_get_current(void)
{
    return current_ss;
}


#define SSH_SET_CURRENT  ssh_session_t *old_ss = ssh_set_current(ss)
#define SSH_RESTORE_CURRENT  ssh_set_current(old_ss)

static void proxy_to64frombits(unsigned char *out, const unsigned char *in, int inlen) SSH_SEGMENT;

static int ssh_proxy_connect(int sock, const char *lpszDestHost, 
	int nDestPort, const char * lpszProxyAddress, const int nProxyPort, 
	const char * lpszProxyProtocol, BOOL bNeedAuth, const char * lpszUserName,
	const char * lpszPassword) SSH_SEGMENT;
void ssh_readproxy(ssh_session_t *ss) SSH_SEGMENT;


static void proxy_to64frombits(unsigned char *out, const unsigned char *in, int inlen)
{
	const char base64digits[] =
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
	for (; inlen >= 3; inlen -= 3)
	{
		*out++ = base64digits[in[0] >> 2];
		*out++ = base64digits[((in[0] << 4) & 0x30) | (in[1] >> 4)];
		*out++ = base64digits[((in[1] << 2) & 0x3c) | (in[2] >> 6)];
		*out++ = base64digits[in[2] & 0x3f];
		in += 3;
	}
	if (inlen > 0)
	{
		unsigned char fragment;

		*out++ = base64digits[in[0] >> 2];
		fragment = (in[0] << 4) & 0x30;
		if (inlen > 1)
		fragment |= in[1] >> 4;
		*out++ = base64digits[fragment];
		*out++ = (inlen < 2) ? '=' : base64digits[(in[1] << 2) & 0x3c];
		*out++ = '=';
	}
	*out = '\0';
};

//-- connect proxy
#define PROXY_BUF_LEN 256
static int ssh_proxy_connect(int sock, const char *lpszDestHost, int nDestPort, const char * lpszProxyAddress, const int nProxyPort, 
	const char * lpszProxyProtocol, BOOL bNeedAuth, const char * lpszUserName, const char * lpszPassword)
{
	//connect to proxy
	char szBuffer[PROXY_BUF_LEN + 1] = {""};
	int nLen = 0;

	NetSocketAddrINType saProxy;

	saProxy.family = AF_INET;
	saProxy.port = nProxyPort;
	saProxy.addr = inet_addr(lpszProxyAddress);
	if (saProxy.addr == INADDR_NONE)
	{
		NetHostInfoBufType *lphost;
		lphost = (NetHostInfoBufType *)gethostbyname(lpszProxyAddress);
		if (lphost != NULL)
			saProxy.addr = lphost->address[0];
		else
		{
			printf("gethostbyname %s err\n",lpszProxyAddress);
			return -1;
		}
	} 

	if(connect(sock, (NetSocketAddrType*)&saProxy, sizeof(saProxy)))
	{
		printf("connect %s err\n",lpszProxyAddress);
		return -1;
	}

	if(strcmp(lpszProxyProtocol, "HTTPS") == 0 || strcmp(lpszProxyProtocol, "HTTP") == 0 )
	{
		if(bNeedAuth)
		{
			//Proxy-Authorization: Basic 
			char szAuth[PROXY_BUF_LEN + 1] = {""};
			char szAuthT[PROXY_BUF_LEN + 1] = {""};
			sprintf(szAuthT, "%s:%s", lpszUserName,lpszPassword);
			proxy_to64frombits((unsigned char *)szAuth, (unsigned char *)szAuthT, strlen(szAuthT));
			sprintf(szBuffer, "CONNECT %s:%d HTTP/1.0\r\nProxy-Authorization: Basic %s\r\n\r\n", lpszDestHost, nDestPort, szAuth);
			printf(szBuffer);
		}
		else
		{
			sprintf(szBuffer, "CONNECT %s:%d HTTP/1.0\r\n\r\n", lpszDestHost, nDestPort);
			printf(szBuffer);
		}

		nLen = strlen(szBuffer);
		send(sock, szBuffer, nLen, 0);
		memset(szBuffer,0,PROXY_BUF_LEN);
		
		nLen = recv(sock, szBuffer, nLen, 0);
		if (nLen < 0) {
			NetworkError("recv data err",errno);
			return -1;
		}
		
		if (nLen == 0) {
			// socket closed
			NetworkError("socket closed",errno);
			return -1;
		}
		

		if(strncmp(szBuffer, "HTTP/1.0 200", strlen("HTTP/1.0 200")) == 0 || 
		strncmp(szBuffer, "HTTP/1.1 200", strlen("HTTP/1.1 200")) == 0)
		{
			printf("connected to proxy ok!!\n");
			return 0;
		}
		return -1;
	}

	if(strcmp(lpszProxyProtocol, "SOCKS5") == 0)
	{
		//auth
		unsigned short uPort;
		nLen = 0;
		if(bNeedAuth)
		{
			szBuffer[0] = 5;
			szBuffer[1] = 2;
			szBuffer[2] = 0;
			szBuffer[3] = 2;
			nLen = 4;
			send(sock, szBuffer, nLen, 0);
			nLen = 2;
			recv(sock, szBuffer, nLen, 0);
			if(szBuffer[0] == 5)
			{
				//need auth
				if(szBuffer[1] == 2)
				{
					szBuffer[0] = 1;
					nLen = strlen(lpszUserName);
					szBuffer[1] = nLen;
					strcpy(szBuffer+2, lpszUserName);
					nLen += 2;
					szBuffer[nLen] = strlen(lpszPassword);
					strcpy(szBuffer+nLen+1, lpszPassword);
					nLen = nLen + 1 + strlen(lpszPassword);
					send(sock, szBuffer, nLen, 0);
					nLen = 2;
					recv(sock, szBuffer, nLen, 0);
					if(szBuffer[1] != 0)
					{
						return -1;
					}
				}
				else
				{
					if(szBuffer[1] != 0)
					{
						return -1;
					}
				}
			}
			else
			return -1;

		}
		else
		{
			szBuffer[0] = 5;
			szBuffer[1] = 1;
			szBuffer[2] = 0;
			nLen = 3;
			send(sock, szBuffer, nLen, 0);
			nLen = 2;
			recv(sock, szBuffer, nLen, 0);
			if(szBuffer[0] != 5 || szBuffer[1] != 0)
			{
				return -1;
			}
		}
		//translate DestAddr
		szBuffer[0] = 5;
		szBuffer[1] = 1;
		szBuffer[2] = 0;
		szBuffer[3] = 3;//DOMAIN
		szBuffer[4] = strlen(lpszDestHost);//domain len
		strcpy(szBuffer+5, lpszDestHost);
		uPort = htons(nDestPort);
		memcpy(szBuffer+5+strlen(lpszDestHost), &uPort, 2);
		nLen = 5 + strlen(lpszDestHost) + 2;
		send(sock, szBuffer, nLen, 0);
		nLen = 10;
		recv(sock, szBuffer, nLen, 0);
		if(szBuffer[0] != 5 || szBuffer[1] != 0)
		{
			return -1;
		}
		return 0;
	}
	return -1;
}




ssh_session_t *ssh_create(char *username, RectangleType bounds)
{
    ssh_session_t *ss = 
        (ssh_session_t *)nonarena_malloc(sizeof(ssh_session_t));
    bzero(ss, sizeof(ssh_session_t));
    
    ssh_set_current(ss);

    ssh_log(__PRETTY_FUNCTION__);

    // nonzero init
    // make session_init() etc. functions if this gets big
    ss->s.socket = -1;

    ss->tty = vt100_new(ss, bounds);
    ss->username = arena_strdup(username);

    ssh_set_current(NULL);

    return ss;
}

// initiate ordinary connection shutdown
// The connection may not actually be closed when this function 
// terminates, and this session must still be processed in the event loop.
// Use ssh_is_closed() to determine when shutdown is complete.
// Use ssh_kill() to immediately close a connection (not cleanly)
void ssh_close(ssh_session_t *ss)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    ss->closing = 1;
	if(ss->protocol == protocolSsh2)
	{
	    if (connection_is_open(ss)) connection_close(ss);
	    if (connection_is_closing(ss)) goto done;
	    
	    if (transport_is_open(ss)) transport_close(ss);
	    if (transport_is_closing(ss)) goto done;

	    if (packetizer_is_open(ss)) packetizer_close(ss);
	    if (packetizer_is_closing(ss)) goto done;
	}
	else if(ss->protocol == protocolTelnet)
	{
	    if (tel_is_open(ss)) tel_close(ss);
	}
    if (session_is_open(ss)) session_close(ss);
    if (session_is_closing(ss)) goto done;

 done:
    SSH_RESTORE_CURRENT;
}

void ssh_kill(ssh_session_t *ss)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    ss->closing = 1;
	if(ss->protocol == protocolSsh2)
	{
		connection_kill(ss);
		transport_kill(ss);
		packetizer_kill(ss);
	}

    session_kill(ss);

    SSH_RESTORE_CURRENT;
}

void ssh_free(ssh_session_t *ss)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    ssh_kill(ss);
    arena_free(ss->username);
    vt100_free(ss->tty);

    SSH_RESTORE_CURRENT;

    // ss itself is not arena allocated
    nonarena_free(ss);
}

void ssh_readproxy(ssh_session_t *ss)
{
	char *str;
    ss->proxy.useproxy = PrefsGetInt(prefUseProxy, 0);

	str = PrefsGetString(prefProxyServer, "10.0.0.172");
	strcpy(ss->proxy.proxyname ,str);
	arena_free(str);
	
	str = PrefsGetString(prefProxyPort, "80");
	strcpy(ss->proxy.proxyport,str);
	arena_free(str);
	
	str = PrefsGetString(prefProxyUser, "cmwap");
	strcpy(ss->proxy.proxyuser,str);
	arena_free(str);
	
	str = PrefsGetString(prefProxyPsw, "cmwap");
	strcpy(ss->proxy.proxypsw,str);
	arena_free(str);
}

int ssh_open(ssh_session_t *ss, char *protocol, char *hostname, int port)
{
    char on = 1;
    int sock;
    int err = 0;
	
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    // find remote host
    // GrP gethostbyname() is supposed to return NULL when the host is not 
    // found, but this doesn't seem to happen on my Tungsten C running 5.2.1.
    // Instead, it returns &AppHostInfo, which may still have a *previous* 
    // host lookup result! 
    h_errno = 0;
    bzero(&AppHostInfo, sizeof(AppHostInfo));

	if(0 == strncmp(protocol,"telnet",strlen("telnet")) )
	{
		ss->protocol = protocolTelnet;
	}
	else if(0 == strncmp(protocol,"ssh2",strlen("ssh2")) )
	{
		ss->protocol = protocolSsh2;
	}
	else
	{
		err = -1; 
		goto done;
	}
	
	ssh_readproxy(ss);
// create socket
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0) { 
		NetworkError("Could not create network socket.", errno);
		err = -1; 
		goto done;
	}
	
	
	// begin connecting to remote host
	// fixme need non-blocking connect
	
	//not use proxy
    if(!ss->proxy.useproxy)
	{
	
		NetHostInfoBufType *remoteHost;
		NetSocketAddrINType remoteINAddr;
		
		remoteHost = (NetHostInfoBufType *)gethostbyname(hostname);
		if (!remoteHost  ||  h_errno  ||  remoteHost->hostInfo.addrLen == 0) { 
			NetworkError("The server could not be found.", h_errno);
			err = -1; 
			goto done; 
		}
		remoteINAddr.family = AF_INET;
		remoteINAddr.port = port;
		printf("port:%d\n",port);
		remoteINAddr.addr = remoteHost->address[0];
		err = connect(sock, (NetSocketAddrType *)&remoteINAddr, sizeof(remoteINAddr));
		
		if (err) { 
			NetworkError("Could not connect to server.", errno);
			goto done;
		}
		
		ss->s.hostaddr = arena_strdup(inet_ntoa(*(struct in_addr *)&remoteINAddr.addr));
		printf("host :%s\n",ss->s.hostaddr);
	}
	else
	{
		err = ssh_proxy_connect(sock,hostname,port,ss->proxy.proxyname,
			atoi(ss->proxy.proxyport),"HTTP",1,ss->proxy.proxyuser,ss->proxy.proxypsw);
		if (err) { 
			NetworkError("proxy Could not connect to server.",  errno);
			goto done;
		}
		ss->s.hostaddr = arena_strdup("10.0.0.172");
	}
	
	// set socket options: non-blocking, no coalescing on send
    setsockopt(sock, SOL_SOCKET, netSocketOptSockNonBlocking, &on, sizeof(on));
    setsockopt(sock, netSocketOptLevelTCP, netSocketOptTCPNoDelay, &on, sizeof(on));

    ss->s.socket = sock;
    ss->s.hostname = arena_strdup(hostname);
    ss->s.state = SESSION_STATE_OPEN;

    // fixme move some of the above to session.c
	if(ss->protocol == protocolSsh2)
	    packetizer_start(ss, "SSH-2.0-pssh_" PSSH_UVERSION);

 done:
    SSH_RESTORE_CURRENT;
    if(sock >=0)close(sock);
    return err;
}


int ssh_is_closed(ssh_session_t *ss)
{
    // fixme
    return (!ss  ||  ss->s.socket == -1);
}


int ssh_request_select(ssh_session_t *ss, fd_set *rfds, fd_set *wfds)
{
    // fixme
    if (ss  &&  ss->s.socket >= 0) {
        FD_SET(ss->s.socket, rfds);
        return ss->s.socket;
    } else {
        return 0;
    }
}


int ssh_is_selected(ssh_session_t *ss, fd_set *fds)
{
    if (ss  &&  ss->s.socket >= 0) {
        return FD_ISSET(ss->s.socket, fds);
    } else {
        return 0;
    }
}


void ssh_read(ssh_session_t *ss)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);
	if (!ss) return;
	if(ss->protocol == protocolSsh2)
	{
		packetizer_receive_data(ss);
	}
	else if(ss->protocol == protocolTelnet)
	{
		if (ss && session_is_open(ss)) 
		{
			tel_receive_data(ss);
		}
	}

    SSH_RESTORE_CURRENT;
}


void ssh_write(ssh_session_t *ss)
{
    // fixme 
    return;
}

int ssh_app_cursor_keys(ssh_session_t *ss)
{
    ssh_log(__PRETTY_FUNCTION__);
    if (ss  &&  ss->tty) return vt100_app_cursor_keys(ss->tty);
    else return 0;
}

void ssh_key(ssh_session_t *ss, char c)
{
    ssh_keys(ss, &c, 1);
}

void ssh_keys(ssh_session_t *ss, uint8_t *buf, uint16_t bufLen)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);
    if (!ss) return;
	if(ss->protocol == protocolSsh2)
	{
	    if (connection_is_open(ss)) {
	        connection_send_data(ss, buf, bufLen);
	    }
	}
	else if(ss->protocol == protocolTelnet)
	{
		if(tel_is_open(ss) )
		{
			tel_send_data(ss,buf,bufLen);
		}
	}
    // maybe scroll to bottom when key is pressed
    vt100_seen_key(ss->tty);

    SSH_RESTORE_CURRENT;
}

void ssh_bytes(ssh_session_t *ss, uint8_t *buf, uint16_t bufLen)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);
    if (!ss) return;
	if(ss->protocol == protocolSsh2)
	{
		if (connection_is_open(ss)) {
		connection_send_data(ss, buf, bufLen);
		}
	}
	else if(ss->protocol == protocolTelnet)	// Non-interactive input - do NOT scroll to bottom
	{
		if(tel_is_open(ss) )
		{
			tel_send_data(ss,buf,bufLen);
		}
	}
    SSH_RESTORE_CURRENT;
}


void ssh_receive_bytes(ssh_session_t *ss, uint8_t *buf, uint16_t bufLen)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__ " %u bytes", bufLen);
    if (!ss) return;

    if (ss->tty) {
        EventType e = {0};
        vt100_write(ss->tty, buf, bufLen);
        e.eType = usrDrawVT100Event;
        EvtAddUniqueEventToQueue(&e, 0, true);
    }

    SSH_RESTORE_CURRENT;
}

void ssh_update(ssh_session_t *ss)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);
    if (!ss) return;

    if (ss->tty) vt100_update(ss->tty);

    ssh_log(__PRETTY_FUNCTION__ " done");

    SSH_RESTORE_CURRENT;
}


void ssh_task(ssh_session_t *ss)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    if (ss  &&  ss->tty) vt100_task(ss->tty);

    ssh_log(__PRETTY_FUNCTION__ " done");

    SSH_RESTORE_CURRENT;
}

void ssh_request_password(int numAttempts)
{
    // do NOT call ssh_use_password until next event loop!
    FrmPopupForm(PasswordFormID);
}

void ssh_use_password(ssh_session_t *ss, char *username, char *password)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);
	if(ss->protocol == protocolSsh2)
	    connection_use_password(ss, username, password);

    SSH_RESTORE_CURRENT;
}


void ssh_request_passphrase(queue_t *q)
{
    // do NOT call ssh_use_password until next event loop!
    PublicKeyChoiceKeys = q;
    FrmPopupForm(PublicKeyChoiceFormID);
}


Boolean ssh_use_passphrase(ssh_session_t *ss, char *username, char *passphrase,
                           MemHandle pubkey)
{
    Boolean ok;

    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);
	if(ss->protocol == protocolSsh2)
	    ok = connection_use_passphrase(ss, username, passphrase, pubkey);

    SSH_RESTORE_CURRENT;

    return ok;
}


void ssh_request_kbdint(char *name, char *instructions, 
                        prompt_t *prompts, int prompt_count)
{
    if (prompt_count == 0) {
        KbdInt0GetResponses(name, instructions);
    } else {
        KbdIntGetResponses(name, instructions, prompts, prompt_count);
    }
}


void ssh_use_kbdint(struct ssh_session_t *ss, char **responses) 
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);
	
	if(ss->protocol == protocolSsh2)
	    connection_use_kbdint(ss, responses);
    SSH_RESTORE_CURRENT;
}


void ssh_scroll(ssh_session_t *ss, int lines)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    if (ss  &&  ss->tty) vt100_scroll(ss->tty, lines);

    SSH_RESTORE_CURRENT;
}

int ssh_visible_height(ssh_session_t *ss) 
{
    int height = 0;
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    if (ss  &&  ss->tty) {
        vt100_size(ss->tty, NULL, &height, NULL);
    }

    SSH_RESTORE_CURRENT;

    return height;
}

void ssh_activate(ssh_session_t *ss)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    if (ss  &&  ss->tty) {
        EventType e = {0};
        vt100_activate(ss->tty);
        e.eType = usrDrawVT100Event;
        EvtAddUniqueEventToQueue(&e, 0, true);
    }

    SSH_RESTORE_CURRENT;
}

void ssh_deactivate(ssh_session_t *ss)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    if (ss  &&  ss->tty) vt100_deactivate(ss->tty);

    SSH_RESTORE_CURRENT;
}

void ssh_set_bounds(ssh_session_t *ss, RectangleType bounds) 
{
    int charsWide, charsHigh;

    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    if (ss  &&  ss->tty) {
        // bounds is LO-RES and entire gadget area
        vt100_set_bounds(ss->tty, bounds);
        
        vt100_size(ss->tty, &charsWide, &charsHigh, &bounds);
        // bounds is now HI-RES and text area only
		if(ss->protocol == protocolSsh2)
	        connection_use_bounds(ss, bounds, charsWide, charsHigh);
    }

    SSH_RESTORE_CURRENT;
}

void ssh_click(ssh_session_t *ss, int x, int y, int clickCount)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    if (ss  &&  ss->tty) vt100_click(ss->tty, x, y, clickCount);

    SSH_RESTORE_CURRENT;
}

void ssh_deselect(ssh_session_t *ss)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    if (ss  &&  ss->tty) vt100_deselect(ss->tty);

    SSH_RESTORE_CURRENT;
}



void ssh_copy(ssh_session_t *ss)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    if (ss  &&  ss->tty) vt100_copy(ss->tty);

    SSH_RESTORE_CURRENT;
}


void ssh_paste(ssh_session_t *ss)
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    if (ss  &&  ss->tty) vt100_paste(ss->tty);

    SSH_RESTORE_CURRENT;
}


int ssh_selection_exists(struct ssh_session_t *ss) 
{
    int result;

    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    if (ss  &&  ss->tty) result = vt100_selection_exists(ss->tty);
    else result = false;

    SSH_RESTORE_CURRENT;

    return result;
}

void ssh_size(struct ssh_session_t *ss, int *charsWide, int *charsHigh) 
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);
    
    if (ss  &&  ss->tty) vt100_size(ss->tty, charsWide, charsHigh, NULL);
    else { *charsWide = 80; *charsHigh = 40; }

    SSH_RESTORE_CURRENT;
}


// current set of mutable prefs:
// * font and font size
// * font colors
void ssh_reread_prefs(struct ssh_session_t *ss) 
{
    SSH_SET_CURRENT;

    ssh_log(__PRETTY_FUNCTION__);

    if (ss && ss->tty) {
        int charsWide, charsHigh;
        RectangleType bounds;

        vt100_reread_prefs(ss->tty);

        // send new size notification in case it changed
        vt100_size(ss->tty, &charsWide, &charsHigh, &bounds);
		if(ss->protocol == protocolSsh2)
	        connection_use_bounds(ss, bounds, charsWide, charsHigh);
    }

    SSH_RESTORE_CURRENT;
}
